"""Integration tests for the HyperTransformer."""

import re

import numpy as np
import pandas as pd
import pytest

from rdt import HyperTransformer, get_demo
from rdt.errors import ConfigNotSetError, InvalidConfigError, InvalidDataError, NotFittedError
from rdt.transformers import (
    AnonymizedFaker, BaseTransformer, BinaryEncoder, ClusterBasedNormalizer, FloatFormatter,
    FrequencyEncoder, LabelEncoder, OneHotEncoder, RegexGenerator, UniformEncoder,
    UnixTimestampEncoder, get_default_transformer, get_default_transformers)
from rdt.transformers.datetime import OptimizedTimestampEncoder
from rdt.transformers.numerical import GaussianNormalizer
from rdt.transformers.pii.anonymizer import PseudoAnonymizedFaker


class DummyTransformerNumerical(BaseTransformer):

    INPUT_SDTYPE = 'categorical'

    def _fit(self, data):
        pass

    def _transform(self, data):
        return data.astype(float)

    def _reverse_transform(self, data):
        return data.astype(str)


class DummyTransformerNotMLReady(BaseTransformer):

    INPUT_SDTYPE = 'datetime'

    def __init__(self):
        super().__init__()
        self.output_properties = {
            None: {'sdtype': 'datetime', 'next_transformer': FrequencyEncoder()}
        }

    def _fit(self, data):
        pass

    def _transform(self, data):
        # Stringify input data
        return data.astype(str)

    def _reverse_transform(self, data):
        return data.astype('datetime64[ns]')


TEST_DATA_INDEX = [4, 6, 3, 8, 'a', 1.0, 2.0, 3.0]


def get_input_data():
    datetimes = pd.to_datetime([
        '2010-02-01',
        '2010-02-01',
        '2010-01-01',
        '2010-01-01',
        '2010-01-01',
        '2010-02-01',
        '2010-01-01',
        '2010-01-01',
    ])
    data = pd.DataFrame({
        'integer': [1, 2, 1, 3, 1, 4, 2, 3],
        'float': [0.1, 0.2, 0.1, 0.2, 0.1, 0.4, 0.2, 0.3],
        'categorical': ['a', 'a', 'b', 'b', 'a', 'b', 'a', 'a'],
        'bool': [False, False, False, True, False, False, True, False],
        'datetime': datetimes,
        'names': ['Jon', 'Arya', 'Arya', 'Jon', 'Jon', 'Sansa', 'Jon', 'Jon'],
    }, index=TEST_DATA_INDEX)

    return data


def get_transformed_data():
    datetimes = [
        1.264982e+18,
        1.264982e+18,
        1.262304e+18,
        1.262304e+18,
        1.262304e+18,
        1.264982e+18,
        1.262304e+18,
        1.262304e+18
    ]
    return pd.DataFrame({
        'integer': [1., 2., 1., 3., 1., 4., 2., 3.],
        'float': [0.1, 0.2, 0.1, 0.2, 0.1, 0.4, 0.2, 0.3],
        'categorical': [
            0.6056724228102, 0.551035999670618, 0.6747435795337998, 0.9245683344321064,
            0.5791232599393884, 0.9570454751421033, 0.5800536682210967, 0.3183267452542666
        ],
        'bool': [
            0.196209398885917, 0.4301613485619816, 0.02000535472243109, 0.7993073999936193,
            0.6645481177917991, 0.25074590518885065, 0.7772353030710347, 0.46369904088373604
        ],
        'datetime': datetimes,
        'names': [
            0.15112620775650704, 0.857444679914493, 0.7654375186193025, 0.42569016008650984,
            0.30010761543029285, 0.9108473448910603, 0.15922866807030298, 0.3875325513134956
        ]
    }, index=TEST_DATA_INDEX)


def get_reversed_data():
    data = get_input_data()

    return data


class TestHyperTransformer:
    def test_default_inputs(self):
        """Test the HyperTransformer with default parameters.

        This tests that if default parameters are provided to the HyperTransformer,
        the ``default_transformers`` method will be used to determine which
        transformers to use for each field.

        Input:
            - A dataframe with every sdtype.
            - A fixed random seed to guarantee the samle values are null.

        Expected behavior:
            - The transformed data should contain all the ML ready data.
            - The reverse transformed data should be the same as the input.
        """
        # Setup
        datetimes = pd.to_datetime([
            np.nan,
            '2010-02-01',
            '2010-01-01',
            '2010-01-01',
            '2010-01-01',
            '2010-02-01',
            '2010-01-01',
            '2010-01-01',
        ])
        data = pd.DataFrame({
            'integer': [1, 2, 1, 3, 1, 4, 2, 3],
            'float': [0.1, 0.2, 0.1, np.nan, 0.1, 0.4, np.nan, 0.3],
            'categorical': ['a', 'a', np.nan, 'b', 'a', 'b', 'a', 'a'],
            'bool': [False, np.nan, False, True, False, np.nan, True, False],
            'datetime': datetimes,
            'names': ['Jon', 'Arya', 'Arya', 'Jon', 'Jon', 'Sansa', 'Jon', 'Jon'],
        }, index=TEST_DATA_INDEX)

        # Run
        ht = HyperTransformer()
        ht.detect_initial_config(data)
        ht.fit(data)
        transformed = ht.transform(data)
        reverse_transformed = ht.reverse_transform(transformed)

        # Assert
        expected_datetimes = [
            1.263069e+18,
            1.264982e+18,
            1.262304e+18,
            1.262304e+18,
            1.262304e+18,
            1.264982e+18,
            1.262304e+18,
            1.262304e+18
        ]
        expected_transformed = pd.DataFrame({
            'integer': [1., 2., 1., 3., 1., 4., 2., 3.],
            'float': [0.1, 0.2, 0.1, 0.2, 0.1, 0.4, 0.2, 0.3],
            'categorical': [
                0.6056724228102, 0.551035999670618, 0.6415811931779333, 0.9497122229547376,
                0.5791232599393884, 0.9713636500947356, 0.5800536682210967, 0.3183267452542666
            ],
            'bool': [
                0.13080626592394468, 0.6433871161873272, 0.013336903148287393, 0.7993073999936193,
                0.4430320785278661, 0.5835819683962835, 0.7772353030710347, 0.3091326939224907
            ],
            'datetime': expected_datetimes,
            'names': [
                0.15112620775650704, 0.857444679914493, 0.7654375186193025, 0.42569016008650984,
                0.30010761543029285, 0.9108473448910603, 0.15922866807030298, 0.3875325513134956
            ]
        }, index=TEST_DATA_INDEX)
        pd.testing.assert_frame_equal(transformed, expected_transformed)

        reversed_datetimes = pd.to_datetime([
            '2010-01-09',
            '2010-02-01',
            '2010-01-01',
            '2010-01-01',
            '2010-01-01',
            '2010-02-01',
            '2010-01-01',
            '2010-01-01',
        ])
        expected_reversed = pd.DataFrame({
            'integer': [1, 2, 1, 3, 1, 4, 2, 3],
            'float': [0.1, 0.2, 0.1, 0.20000000000000004, 0.1, 0.4, 0.20000000000000004, 0.3],
            'categorical': ['a', 'a', np.nan, 'b', 'a', 'b', 'a', 'a'],
            'bool': [False, False, False, True, False, False, True, False],
            'datetime': reversed_datetimes,
            'names': ['Jon', 'Arya', 'Arya', 'Jon', 'Jon', 'Sansa', 'Jon', 'Jon'],
        }, index=TEST_DATA_INDEX)
        for row in range(reverse_transformed.shape[0]):
            for column in range(reverse_transformed.shape[1]):
                expected = expected_reversed.iloc[row, column]
                actual = reverse_transformed.iloc[row, column]
                assert pd.isna(actual) or expected == actual

        assert isinstance(ht.field_transformers['integer'], FloatFormatter)
        assert isinstance(ht.field_transformers['float'], FloatFormatter)
        assert isinstance(ht.field_transformers['categorical'], UniformEncoder)
        assert isinstance(ht.field_transformers['bool'], UniformEncoder)
        assert isinstance(ht.field_transformers['datetime'], UnixTimestampEncoder)
        assert isinstance(ht.field_transformers['names'], UniformEncoder)

        get_default_transformers.cache_clear()
        get_default_transformer.cache_clear()

    def test_field_transformers(self):
        """Test the HyperTransformer with ``field_transformers`` provided.

        This tests that the transformers specified in the ``field_transformers``
        argument are used. Any output of a transformer that is not ML ready (not
        in the ``_transform_output_sdtypes`` list) should be recursively transformed
        till it is.

        Setup:
            - The datetime column is set to use a dummy transformer that stringifies
            the input. That output is then set to use the categorical transformer.

        Input:
            - A dict mapping each field to a transformer.
            - A dataframe with every sdtype.

        Expected behavior:
            - The transformed data should contain all the ML ready data.
            - The reverse transformed data should be the same as the input.
        """
        # Setup
        config = {
            'sdtypes': {
                'integer': 'numerical',
                'float': 'numerical',
                'categorical': 'categorical',
                'bool': 'boolean',
                'datetime': 'datetime',
                'names': 'categorical'
            },
            'transformers': {
                'integer': FloatFormatter(missing_value_replacement='mean'),
                'float': FloatFormatter(missing_value_replacement='mean'),
                'categorical': UniformEncoder(),
                'bool': UniformEncoder(),
                'datetime': DummyTransformerNotMLReady(),
                'names': UniformEncoder()
            }
        }

        data = get_input_data()

        # Run
        ht = HyperTransformer()
        ht.detect_initial_config(data)
        ht.set_config(config)
        ht.fit(data)
        transformed = ht.transform(data)
        reverse_transformed = ht.reverse_transform(transformed)

        # Assert
        expected_transformed = get_transformed_data()
        transformed_datetimes = [0.8125, 0.8125, 0.3125, 0.3125, 0.3125, 0.8125, 0.3125, 0.3125]
        expected_transformed['datetime'] = transformed_datetimes
        pd.testing.assert_frame_equal(transformed, expected_transformed)

        expected_reversed = get_reversed_data()
        pd.testing.assert_frame_equal(expected_reversed, reverse_transformed)

    def test_single_category(self):
        """Test that categorical variables with a single value are supported."""
        # Setup
        ht = HyperTransformer()
        data = pd.DataFrame({
            'a': ['a', 'a', 'a']
        })

        # Run
        ht.detect_initial_config(data)
        ht.update_transformers(column_name_to_transformer={
            'a': OneHotEncoder()
        })
        ht.fit(data)
        transformed = ht.transform(data)
        reverse = ht.reverse_transform(transformed)

        # Assert
        pd.testing.assert_frame_equal(data, reverse)

    def test_categorical_encoders_with_booleans(self):
        """Test that categorical encoders support boolean values."""
        # Setup
        config = {
            'sdtypes': {
                'email_confirmed': 'boolean',
                'subscribed': 'boolean',
                'paid': 'boolean',
            },
            'transformers': {
                'email_confirmed': FrequencyEncoder(),
                'subscribed': OneHotEncoder(),
                'paid': LabelEncoder(),
            }
        }

        ht = HyperTransformer()

        # Run and Assert
        ht.set_config(config)

    def test_dtype_category(self):
        """Test that categorical variables of dtype category are supported."""
        # Setup
        data = pd.DataFrame({'a': ['a', 'b', 'c']}, dtype='category')

        # Run
        ht = HyperTransformer()
        ht.detect_initial_config(data)
        ht.fit(data)
        transformed = ht.transform(data)
        reverse = ht.reverse_transform(transformed)

        # Assert
        pd.testing.assert_frame_equal(reverse, data)

    def test_multiple_fits(self):
        """HyperTransformer should be able to be used multiple times.

        Fitting, transforming and reverse transforming should produce the same results when
        called on the same data multiple times.
        """
        # Setup
        data = get_input_data()
        ht = HyperTransformer()

        # Run
        ht.detect_initial_config(data)
        ht.fit(data)
        transformed1 = ht.transform(data)
        reversed1 = ht.reverse_transform(transformed1)

        ht.detect_initial_config(data)
        ht.fit(data)
        transformed2 = ht.transform(data)
        reversed2 = ht.reverse_transform(transformed2)

        # Assert
        pd.testing.assert_frame_equal(transformed1, transformed2)
        pd.testing.assert_frame_equal(reversed1, reversed2)

    def test_multiple_fits_different_data(self):
        """HyperTransformer should be able to be used multiple times regardless of the data.

        Fitting, transforming and reverse transforming should work when called on different data.
        """
        # Setup
        data = pd.DataFrame({'col1': [1, 2, 3], 'col2': [1.0, 0.0, 0.0]})
        new_data = pd.DataFrame({'col2': [1, 2, 3], 'col1': [1.0, 0.0, 0.0]})
        ht = HyperTransformer()

        # Run
        ht.detect_initial_config(data)
        ht.fit(data)
        ht.detect_initial_config(new_data)
        ht.fit(new_data)
        transformed1 = ht.transform(new_data)
        transformed2 = ht.transform(new_data)
        reverse1 = ht.reverse_transform(transformed1)
        reverse2 = ht.reverse_transform(transformed2)

        # Assert
        expected_transformed = pd.DataFrame(
            {'col2': [1., 2., 3.], 'col1': [1.0, 0.0, 0.0]})
        pd.testing.assert_frame_equal(transformed1, expected_transformed)
        pd.testing.assert_frame_equal(transformed2, expected_transformed)
        pd.testing.assert_frame_equal(reverse1, new_data)
        pd.testing.assert_frame_equal(reverse2, new_data)

    def test_multiple_fits_different_columns(self):
        """HyperTransformer should be able to be used multiple times regardless of the data.

        Fitting, transforming and reverse transforming should work when called on different data.
        """
        # Setup
        data = pd.DataFrame({'col1': [1, 2, 3], 'col2': [1.0, 0.0, 0.0]})
        new_data = pd.DataFrame({'col3': [1, 2, 3], 'col4': [1.0, 0.0, 0.0]})
        ht = HyperTransformer()

        # Run
        ht.detect_initial_config(data)
        ht.fit(data)
        ht.detect_initial_config(new_data)
        ht.fit(new_data)
        transformed1 = ht.transform(new_data)
        transformed2 = ht.transform(new_data)
        reverse1 = ht.reverse_transform(transformed1)
        reverse2 = ht.reverse_transform(transformed2)

        # Assert
        expected_transformed = pd.DataFrame(
            {'col3': [1., 2., 3.], 'col4': [1.0, 0.0, 0.0]})
        pd.testing.assert_frame_equal(transformed1, expected_transformed)
        pd.testing.assert_frame_equal(transformed2, expected_transformed)
        pd.testing.assert_frame_equal(reverse1, new_data)
        pd.testing.assert_frame_equal(reverse2, new_data)

    def test_multiple_fits_with_set_config(self):
        """HyperTransformer should be able to be used multiple times regardless of the data.

        Fitting, transforming and reverse transforming should work when called on different data.
        """
        # Setup
        data = get_input_data()
        ht = HyperTransformer()

        # Run
        ht.detect_initial_config(data)
        ht.set_config(config={
            'sdtypes': {'integer': 'categorical'},
            'transformers': {'integer': FrequencyEncoder()}
        })
        ht.fit(data)
        transformed1 = ht.transform(data)
        reverse1 = ht.reverse_transform(transformed1)

        ht.fit(data)
        transformed2 = ht.transform(data)
        reverse2 = ht.reverse_transform(transformed2)

        # Assert
        pd.testing.assert_frame_equal(transformed1, transformed2)
        pd.testing.assert_frame_equal(reverse1, reverse2)

    def test_multiple_detect_configs_with_set_config(self):
        """HyperTransformer should be able to be used multiple times regardless of the data.

        Fitting, transforming and reverse transforming should work when called on different data.
        """
        # Setup
        data = get_input_data()
        ht = HyperTransformer()

        # Run
        ht.detect_initial_config(data)
        ht.fit(data)
        transformed1 = ht.transform(data)
        reverse1 = ht.reverse_transform(transformed1)

        ht.set_config(config={
            'sdtypes': {'integers': 'categorical'},
            'transformers': {'integers': FrequencyEncoder()}
        })

        ht.detect_initial_config(data)
        ht.fit(data)
        transformed2 = ht.transform(data)
        reverse2 = ht.reverse_transform(transformed2)

        # Assert
        pd.testing.assert_frame_equal(transformed1, transformed2)
        pd.testing.assert_frame_equal(reverse1, reverse2)

    def test_detect_initial_config_doesnt_affect_fit(self):
        """HyperTransformer should fit the same way regardless of ``detect_initial_config``.

        Calling the ``detect_initial_config`` method should not affect the results of ``fit``,
        ``transform`` or ``reverse_transform``.
        """
        # Setup
        data = get_input_data()
        ht = HyperTransformer()

        # Run
        ht.detect_initial_config(data)
        ht.fit(data)
        transformed1 = ht.transform(data)
        reversed1 = ht.reverse_transform(transformed1)

        ht.detect_initial_config(data)
        ht.fit(data)
        transformed2 = ht.transform(data)
        reversed2 = ht.reverse_transform(transformed1)

        # Assert
        pd.testing.assert_frame_equal(transformed1, transformed2)
        pd.testing.assert_frame_equal(reversed1, reversed2)

    def test_multiple_detects(self):
        """HyperTransformer should be able to be used multiple times regardless of the data.

        Fitting, transforming and reverse transforming should work when called on different data.
        """
        # Setup
        data = pd.DataFrame({'col2': [1, 2, 3], 'col1': [1.0, 0.0, 0.0]})
        new_data = get_input_data()
        ht = HyperTransformer()

        # Run
        ht.detect_initial_config(data)
        ht.detect_initial_config(new_data)
        ht.fit(new_data)
        transformed = ht.transform(new_data)
        reverse = ht.reverse_transform(transformed)

        # Assert
        pd.testing.assert_frame_equal(transformed, get_transformed_data())
        pd.testing.assert_frame_equal(reverse, get_reversed_data())

    def test_transform_without_fit(self):
        """HyperTransformer should raise an error when transforming without fitting."""
        # Setup
        data = pd.DataFrame({'column': [1, 2, 3]})
        ht = HyperTransformer()
        ht.detect_initial_config(data)

        # Run / Assert
        with pytest.raises(NotFittedError):
            ht.transform(data)

    def test_fit_data_different_than_detect(self):
        """HyperTransformer should raise an error when transforming without fitting."""
        # Setup
        ht = HyperTransformer()
        detect_data = pd.DataFrame({'col1': [1, 2], 'col2': ['a', 'b']})
        data = pd.DataFrame({'col1': [1, 2], 'col3': ['a', 'b']})

        # Run / Assert
        error_msg = re.escape(
            'The data you are trying to fit has different columns than the original '
            "detected data (unknown columns: ['col3']). Column names and their "
            "sdtypes must be the same. Use the method 'get_config()' to see the expected "
            'values.'
        )
        ht.detect_initial_config(detect_data)
        with pytest.raises(InvalidDataError, match=error_msg):
            ht.fit(data)

    def test_transform_without_fitting(self):
        """HyperTransformer shouldn't transform when fit hasn't been called yet."""
        # Setup
        data = pd.DataFrame({'col1': [1, 2], 'col2': ['a', 'b']})
        ht = HyperTransformer()

        # Run / Assert
        ht.detect_initial_config(data)
        error_msg = (
            'The HyperTransformer is not ready to use. Please fit your data first using '
            "'fit' or 'fit_transform'."
        )
        with pytest.raises(NotFittedError, match=error_msg):
            ht.transform(data)

    def test_transform_without_refitting(self):
        """HyperTransformer shouldn't transform when a new config hasn't been fitted."""
        # Setup
        data = pd.DataFrame({'col1': [1, 2], 'col2': ['a', 'b']})
        ht = HyperTransformer()

        # Run / Assert
        ht.detect_initial_config(data)
        ht.fit(data)
        ht.update_sdtypes({'col1': 'categorical'})
        error_msg = (
            'The HyperTransformer is not ready to use. Please fit your data first using '
            "'fit' or 'fit_transform'."
        )
        with pytest.raises(NotFittedError, match=error_msg):
            ht.transform(data)

    def test_transform_without_config(self):
        """HyperTransformer shouldn't transform when a config hasn't been set."""
        # Setup
        data = pd.DataFrame({'col1': [1, 2], 'col2': ['a', 'b']})
        ht = HyperTransformer()

        # Run / Assert
        error_msg = (
            "No config detected. Set the config using 'set_config' or pre-populate "
            "it automatically from your data using 'detect_initial_config' prior to "
            'fitting your data.'
        )
        with pytest.raises(ConfigNotSetError, match=error_msg):
            ht.transform(data)

    def test_transform_unseen_columns(self):
        """HyperTransformer shouldn't transform when the data wasn't seen during fit."""
        # Setup
        data = pd.DataFrame({'col1': [1, 2], 'col2': ['a', 'b']})
        different_data = pd.DataFrame({'col3': [1, 2]})
        ht = HyperTransformer()

        # Run / Assert
        ht.detect_initial_config(data)
        ht.fit(data)
        error_msg = error_msg = (
            'The data you are trying to transform has different columns than the original data. '
            'Column names and their sdtypes must be the same.'
        )
        with pytest.raises(InvalidDataError, match=error_msg):
            ht.transform(different_data)

    def test_update_sdtypes_incorrect_columns(self):
        """HyperTransformer should crash when update_sdytpes is passed non-existing columns."""
        # Setup
        data = pd.DataFrame({'col1': [1, 2], 'col2': ['a', 'b']})
        column_name_to_sdtype = {'col3': [1, 2]}
        ht = HyperTransformer()

        # Run / Assert
        ht.detect_initial_config(data)
        error_msg = error_msg = re.escape(
            "Invalid column names: ['col3']. These columns do not exist in the "
            "config. Use 'set_config()' to write and set your entire config at once."
        )
        with pytest.raises(InvalidConfigError, match=error_msg):
            ht.update_sdtypes(column_name_to_sdtype)

    def test_update_sdtypes_incorrect_sdtype(self):
        """HyperTransformer should crash when update_sdytpes is passed non-existing columns."""
        # Setup
        data = pd.DataFrame({'col1': [1, 2], 'col2': ['a', 'b']})
        column_name_to_sdtype = {'col1': 'unexpected'}
        ht = HyperTransformer()

        # Run / Assert
        ht.detect_initial_config(data)
        error_msg = error_msg = re.escape(
            "Invalid sdtypes: ['unexpected']. If you are trying to use a "
            'premium sdtype, contact info@sdv.dev about RDT Add-Ons.'
        )
        with pytest.raises(InvalidConfigError, match=error_msg):
            ht.update_sdtypes(column_name_to_sdtype)

    def test_transform_subset(self):
        """Test the ``transform_subset`` method.

        The method should return a ``pandas.DataFrame`` with the subset of columns transformed.

        Setup:
            - Detect the config and fit the data.

        Input:
            - A ``pandas.DataFrame`` with a subset of the fitted columns.

        Ouput:
            - A ``pandas.DataFrame`` with the subset transformed.
        """
        # Setup
        data = pd.DataFrame({'col1': [1, 2], 'col2': ['a', 'b']})
        subset = data[['col1']]
        ht = HyperTransformer()
        ht.detect_initial_config(data)
        ht.fit(data)

        # Run
        transformed = ht.transform_subset(subset)

        # Assert
        expected = pd.DataFrame({'col1': [1., 2.]})
        pd.testing.assert_frame_equal(transformed, expected)

    def test_reverse_transform_subset(self):
        """Test the ``reverse_transform_subset`` method.

        The method should return a ``pandas.DataFrame`` with the subset of columns reverse
        transformed.

        Setup:
            - Detect the config and fit the data.

        Input:
            - A ``pandas.DataFrame`` with a subset of the output columns.

        Ouput:
            - A ``pandas.DataFrame`` with the subset reverse transformed.
        """
        # Setup
        data = pd.DataFrame({'col1': [1, 2], 'col2': ['a', 'b']})
        subset = pd.DataFrame({'col1': [1, 2]})
        ht = HyperTransformer()
        ht.detect_initial_config(data)
        ht.fit(data)

        # Run
        reverse_transformed = ht.reverse_transform_subset(subset)

        # Assert
        expected = pd.DataFrame({'col1': [1, 2]})
        pd.testing.assert_frame_equal(reverse_transformed, expected)

    def test_with_multiple_supported_sdtypes(self):
        """Test the ``HyperTransformer`` supports multiple ``sdtypes`` for a ``Transformer``.

        Test that the ``HyperTransformer`` works with ``get_supported_sdtypes`` allowing us
        to asign different transformer to a ``sdtype``. For example, a ``FrequencyEncoder`` to
        a ``boolean`` sdtype.

        Setup:
            - Dataframe with multiple datatypes.
            - Instance of ``HyperTransformer``.
            - Update the transformer for ``boolean`` sdtype to ``FrequencyEncoder()``.

        Run:
            - Run end to end the ``hypertransformer``.

        Assert:
            - Assert that the ``FerquencyEncoder`` is used for the ``boolean`` data.
        """
        # Setup
        data = pd.DataFrame({
            'user': ['John', 'Doe', 'John Doe', 'Doe John'],
            'id': list(range(4)),
            'subscribed': [True, False, True, False]
        })

        ht = HyperTransformer()
        ht.detect_initial_config(data)
        ht.update_transformers_by_sdtype(
            sdtype='boolean',
            transformer=FrequencyEncoder(add_noise=True)
        )

        # Run
        transformed = ht.fit_transform(data)
        ht.reverse_transform(transformed)

        for transformer in ht.get_config()['transformers'].values():
            assert not isinstance(transformer, BinaryEncoder)

    def test_reverse_transform_subset_and_generators(self):
        """Test the ``HyperTransformer`` with ``reverse_transform_subset``.

        Test that when calling ``reverse_transform_subset`` and there are ``generators`` like
        ``AnonymizedFaker`` or ``RegexGenerator`` those are not being used in the ``subset``, and
        also any other transformer which can't transform the given columns.

        Setup:
            - DataFrame with multiple datatypes.
            - Instance of HyperTransformer.
            - Add ``pii`` using ``AnonymizedFaker`` and ``RegexGenerator``.

        Run:
            - Use ``fit_transform`` then ``revese_transform_subsample``.

        Assert:
            - Assert that the ``reverse_transformed`` data does not contain any additional columns
            but the expected one.
        """
        # Setup
        customers = get_demo()
        customers['id'] = ['ID_a', 'ID_b', 'ID_c', 'ID_d', 'ID_e']

        # Create a config
        ht = HyperTransformer()
        ht.detect_initial_config(customers)

        # credit_card and id are pii and text columns
        ht.update_sdtypes({
            'credit_card': 'pii',
            'id': 'text'
        })

        ht.update_transformers({
            'credit_card': AnonymizedFaker(),
            'id': RegexGenerator(regex_format='id_[a-z]')
        })

        # Run
        ht.fit(customers)
        transformed = ht.transform(customers)
        reverse_transformed = ht.reverse_transform_subset(transformed[['last_login']])

        # Assert
        expected_transformed_columns = [
            'last_login',
            'email_optin',
            'age',
            'dollars_spent'
        ]
        assert all(expected_transformed_columns == transformed.columns)
        assert reverse_transformed.columns == ['last_login']

    def test_set_config_with_supported_sdtypes(self):
        """Test when setting config with an sdtype that is supported by the transformer."""
        # Setup
        config = {
            'transformers': {
                'boolean_col': FrequencyEncoder(add_noise=True),
            },
            'sdtypes': {
                'boolean_col': 'boolean'
            }
        }
        ht = HyperTransformer()

        # Run and Assert
        ht.set_config(config)

    def test_chained_transformers(self):
        """Test when a transformer is chained to another transformer.

        When the specified transformer indicates a next transformer, they should each be applied in
        order during the transform step, and then reversed during the reverse_transform.
        """
        # Setup
        class DoublingTransformer(BaseTransformer):
            INPUT_SDTYPE = 'numerical'

            def _fit(self, data):
                ...

            def _transform(self, data):
                return data * 2

            def _reverse_transform(self, data):
                return data / 2

        transformer3 = DoublingTransformer()
        transformer2 = DoublingTransformer()
        transformer2.output_properties[None]['next_transformer'] = transformer3
        transformer1 = DoublingTransformer()
        transformer1.output_properties[None]['next_transformer'] = transformer2

        ht = HyperTransformer()
        data = pd.DataFrame({'col': [1., 2, -1, 3, 1]})

        # Run and Assert
        ht.set_config({
            'sdtypes': {'col': 'numerical'},
            'transformers': {'col': transformer1}
        })
        ht.fit(data)

        transformed = ht.transform(data)
        expected_transform = pd.DataFrame({'col': [8., 16, -8, 24, 8]})
        pd.testing.assert_frame_equal(transformed, expected_transform)

        reverse_transformed = ht.reverse_transform(transformed)
        pd.testing.assert_frame_equal(reverse_transformed, data)

    def test_chained_transformers_various_transformers(self):
        """Test when a transformer is chained to another transformer.

        When the specified transformer indicates a next transformer, they should each be applied in
        order during the transform step, and then reversed during the reverse_transform.
        """
        # Setup
        class AB(BaseTransformer):
            INPUT_SDTYPE = 'categorical'

            def _fit(self, data):
                self.output_properties = {
                    None: {'sdtype': 'categorical', 'next_transformer': None},
                    'a': {'sdtype': 'categorical', 'next_transformer': CD()},
                    'b': {'sdtype': 'categorical', 'next_transformer': None},
                }

            def _transform(self, data):
                new_data = pd.DataFrame({f'{self.column_prefix}': data})
                new_data[f'{self.column_prefix}.a'] = data + 'a'
                new_data[f'{self.column_prefix}.b'] = data + 'b'
                return new_data

            def _reverse_transform(self, data):
                new_data = pd.DataFrame()
                new_data[f'{self.column_prefix}'] = data[f'{self.column_prefix}.a'].str[:-1]
                return new_data

        class CD(BaseTransformer):
            INPUT_SDTYPE = 'categorical'

            def _fit(self, data):
                self.output_properties = {
                    'c': {'sdtype': 'categorical', 'next_transformer': None},
                    'd': {'sdtype': 'categorical', 'next_transformer': E()}
                }

            def _transform(self, data):
                new_data = pd.DataFrame()
                new_data[f'{self.column_prefix}.c'] = data + 'c'
                new_data[f'{self.column_prefix}.d'] = data + 'd'
                return new_data

            def _reverse_transform(self, data):
                new_data = pd.DataFrame()
                new_data[f'{self.column_prefix}'] = data[f'{self.column_prefix}.c'].str[:-1]
                return new_data

        class E(BaseTransformer):
            INPUT_SDTYPE = 'categorical'

            def _fit(self, data):
                self.output_properties = {
                    None: {'sdtype': 'categorical', 'next_transformer': None},
                    'e': {'sdtype': 'categorical', 'next_transformer': None}
                }

            def _transform(self, data):
                new_data = pd.DataFrame({f'{self.column_prefix}': data})
                new_data[f'{self.column_prefix}.e'] = data + 'e'
                return new_data

            def _reverse_transform(self, data):
                new_data = pd.DataFrame()
                new_data[f'{self.column_prefix}'] = data[f'{self.column_prefix}.e'].str[:-1]
                return new_data

        ht = HyperTransformer()
        data = pd.DataFrame({
            'col': ['a', 'b', 'c'],
            'col.a': ['1', '2', '3'],
            'col#': ['_', '_', '_']
        })

        # Run and Assert
        ht.set_config({
            'sdtypes': {'col': 'categorical', 'col.a': 'categorical', 'col#': 'categorical'},
            'transformers': {'col': AB(), 'col.a': AB(), 'col#': E()}
        })
        ht.fit(data)
        transformed = ht.transform(data)
        expected = pd.DataFrame({
            'col##': ['a', 'b', 'c'],
            'col##.a.c': ['aac', 'bac', 'cac'],
            'col##.a.d': ['aad', 'bad', 'cad'],
            'col##.a.d.e': ['aade', 'bade', 'cade'],
            'col##.b': ['ab', 'bb', 'cb'],
            'col.a': ['1', '2', '3'],
            'col.a.a.c': ['1ac', '2ac', '3ac'],
            'col.a.a.d': ['1ad', '2ad', '3ad'],
            'col.a.a.d.e': ['1ade', '2ade', '3ade'],
            'col.a.b': ['1b', '2b', '3b'],
            'col#': ['_', '_', '_'],
            'col#.e': ['_e', '_e', '_e'],
        })
        pd.testing.assert_frame_equal(transformed, expected)

        reverse_transformed = ht.reverse_transform(transformed)
        pd.testing.assert_frame_equal(reverse_transformed, data)

    def test_field_transformers_correctly_set(self):
        """Test field_transformers is correctly set through various methods."""
        # Setup
        ht = HyperTransformer()
        data = pd.DataFrame({'col': ['a', 'b', 'c']})

        # getting detected transformers should give the actual transformer
        ht.detect_initial_config(data)
        transformer = ht.get_config()['transformers']['col']
        transformer.new_attribute = 'abc'
        assert ht.get_config()['transformers']['col'].new_attribute == 'abc'

        # the actual transformer should be fitted
        ht.fit(data)
        transformer.new_attribute2 = '123'
        assert ht.get_config()['transformers']['col'].new_attribute == 'abc'
        assert ht.get_config()['transformers']['col'].new_attribute2 == '123'

        # if a transformer was set, it should use the provided instance
        fe = FrequencyEncoder()
        ht.set_config({'sdtypes': {'col': 'categorical'}, 'transformers': {'col': fe}})
        ht.fit(data)
        transformer = ht.get_config()['transformers']['col']
        assert transformer is fe

        # the three cases below make sure any form of acess to the field_transformers
        # correctly accesses and stores the actual transformers
        fe = FrequencyEncoder()
        ht.update_transformers({'col': fe})
        ht.fit(data)
        transformer = ht.get_config()['transformers']['col']
        assert transformer is fe

        ht.update_transformers_by_sdtype('categorical', transformer_name='FrequencyEncoder')
        transformer = ht.get_config()['transformers']['col']
        transformer.new_attribute3 = 'abc'
        ht.fit(data)
        assert ht.get_config()['transformers']['col'].new_attribute3 == 'abc'

        ht.update_sdtypes({'col': 'text'})
        transformer = ht.get_config()['transformers']['col']
        transformer.new_attribute3 = 'abc'
        ht.fit(data)
        assert ht.get_config()['transformers']['col'].new_attribute3 == 'abc'

    def _get_hyper_transformer_with_random_transformers(self, data):
        ht = HyperTransformer()
        ht.detect_initial_config(data)
        ht.update_sdtypes({
            'credit_card': 'pii',
            'name': 'text',
            'signup_day': 'datetime'
        })
        ht.update_transformers({
            'credit_card': AnonymizedFaker('credit_card', 'credit_card_number'),
            'balance': ClusterBasedNormalizer(max_clusters=3),
            'name': RegexGenerator()
        })
        ht.update_transformers_by_sdtype(
            'categorical',
            transformer_name='FrequencyEncoder',
            transformer_parameters={'add_noise': True}
        )

        return ht

    def test_reset_randomization(self):
        """Test that the random seeds are properly set and reset.

        If two separate ``HyperTransformer``s are fit, they should have the same parameters
        and produce the same data when transforming. Successive transforming calls should
        yield different results.

        For reverse transforming, different values should be seen if randomization is involved
        unless ``reset_randomization`` is called.
        """
        # Setup
        data = pd.DataFrame({
            'credit_card': ['123456789', '987654321', '192837645', '918273465', '123789456'],
            'age': [18, 25, 54, 60, 31],
            'name': ['Bob', 'Jane', 'Jack', 'Jill', 'Joe'],
            'signup_day': ['1/1/2020', np.nan, '4/1/2019', '12/1/2008', '5/16/2016'],
            'balance': [250, 5400, 150000, np.nan, 91000],
            'card_type': ['Visa', 'Visa', 'Master Card', 'Amex', 'Visa']
        })
        ht1 = self._get_hyper_transformer_with_random_transformers(data)
        ht2 = self._get_hyper_transformer_with_random_transformers(data)

        # Test transforming multiple times with different transformers
        expected_first_transformed = pd.DataFrame({
            'age': [18.0, 25.0, 54.0, 60.0, 31.0],
            'signup_day': [1.577837e+18, 1.455840e+18, 1.554077e+18, 1.228090e+18, 1.463357e+18],
            'balance.normalized': [
                -2.693016e-01,
                -2.467182e-01,
                3.873711e-01,
                9.571797e-17,
                1.286486e-01
            ],
            'balance.component': [0.0, 0, 0, 0, 0],
            'card_type': [
                0.31440326318001877,
                0.2879792449993428,
                0.7147347796329043,
                0.9397813187279729,
                0.25144169889394075
            ]
        })
        expected_second_transformed = pd.DataFrame({
            'age': [18.0, 25.0, 54.0, 60.0, 31.0],
            'signup_day': [1.577837e+18, 1.455840e+18, 1.554077e+18, 1.228090e+18, 1.463357e+18],
            'balance.normalized': [
                -2.693016e-01,
                -2.467182e-01,
                3.873711e-01,
                9.571797e-17,
                1.286486e-01
            ],
            'balance.component': [0.0, 0, 0, 0, 0],
            'card_type': [
                0.20248666820805558,
                0.4408301080724041,
                0.7082705433167992,
                0.8911691002937682,
                0.2679993642397502
            ]
        })

        ht1.fit(data)
        first_transformed1 = ht1.transform(data)
        ht2.fit(data)
        first_transformed2 = ht2.transform(data)
        second_transformed1 = ht1.transform(data)

        pd.testing.assert_frame_equal(first_transformed1, expected_first_transformed)
        pd.testing.assert_frame_equal(first_transformed2, expected_first_transformed)
        pd.testing.assert_frame_equal(second_transformed1, expected_second_transformed)

        # test reverse transforming multiple times with different tranformers
        expected_first_reverse = pd.DataFrame({
            'credit_card': [
                '3564483245479407',
                '4863061245886958069',
                '4039466324278480',
                '4217004814656859',
                '4343691397776091'
            ],
            'age': [18, 25, 54, 60, 31],
            'name': ['AAAAA', 'AAAAB', 'AAAAC', 'AAAAD', 'AAAAE'],
            'signup_day': ['01/01/2020', '02/19/2016', '04/01/2019', np.nan, np.nan],
            'balance': [250, 5400, 150000, 61662.5, 91000],
            'card_type': ['Visa', 'Visa', 'Master Card', 'Amex', 'Visa']
        })
        expected_second_reverse = pd.DataFrame({
            'credit_card': [
                '4208002654643',
                '3547584322792794',
                '30187802217181',
                '4138954513622487900',
                '346502559595986'
            ],
            'age': [18, 25, 54, 60, 31],
            'name': ['AAAAF', 'AAAAG', 'AAAAH', 'AAAAI', 'AAAAJ'],
            'signup_day': ['01/01/2020', np.nan, '04/01/2019', '12/01/2008', np.nan],
            'balance': [np.nan, 5400, np.nan, 61662.5, 91000],
            'card_type': ['Visa', 'Visa', 'Master Card', 'Amex', 'Visa']
        })
        first_reverse1 = ht1.reverse_transform(first_transformed1)
        first_reverse2 = ht2.reverse_transform(first_transformed1)
        second_reverse1 = ht1.reverse_transform(first_transformed1)
        pd.testing.assert_frame_equal(first_reverse1, expected_first_reverse)
        pd.testing.assert_frame_equal(first_reverse2, expected_first_reverse)
        pd.testing.assert_frame_equal(expected_second_reverse, second_reverse1)

        # Test resetting randomization
        ht1.reset_randomization()

        transformed_post_reset = ht1.reverse_transform(first_transformed1)
        pd.testing.assert_frame_equal(transformed_post_reset, expected_first_reverse)

    def test_cluster_based_normalizer_randomization(self):
        """Test that the ``ClusterBasedNormalizer`` handles randomization correctly.

        If the ``ClusterBasedNormalizer`` transforms the same data multiple times,
        it may yield different results due to randomness. However, if a new instance is created,
        each matching call should yield the same results (ie. call 1 of the first transformer
        should match call 1 of the second).
        """
        data = get_demo(100)
        ht = HyperTransformer()
        ht.detect_initial_config(data)
        ht.update_transformers({
            'age': ClusterBasedNormalizer()
        })
        ht.fit(data)
        transformed1 = ht.transform(data)
        transformed2 = ht.transform(data)

        assert any(transformed1['age.normalized'] != transformed2['age.normalized'])

        ht2 = HyperTransformer()
        ht2.detect_initial_config(data)
        ht2.update_transformers({
            'age': ClusterBasedNormalizer()
        })
        ht2.fit(data)

        pd.testing.assert_frame_equal(transformed1, ht2.transform(data))

    def test_anonymized_faker(self):
        """Test ``AnonymizedFaker`` generates different random values for different columns.

        Issue: https://github.com/sdv-dev/RDT/issues/619.
        """
        # Setup
        data = pd.DataFrame({
            'id1': ['a', 'b', 'c'],
            'id2': ['d', 'e', 'f'],
        })
        ht = HyperTransformer()

        # Run - simple run
        ht.detect_initial_config(data)
        ht.update_sdtypes({
            'id1': 'pii',
            'id2': 'pii'
        })
        ht.update_transformers({
            'id1': AnonymizedFaker(),
            'id2': AnonymizedFaker()
        })
        ht.fit(data)
        transformed = ht.transform(data)
        reverse_transformed1 = ht.reverse_transform(transformed)

        # Assert
        assert reverse_transformed1['id1'].tolist() != reverse_transformed1['id2'].tolist()

        # Run - make sure transforming again returns different values than the original transform
        transformed = ht.transform(data)
        reverse_transformed2 = ht.reverse_transform(transformed)

        # Assert
        assert reverse_transformed2['id1'].tolist() != reverse_transformed2['id2'].tolist()
        assert reverse_transformed1['id1'].tolist() != reverse_transformed2['id1'].tolist()
        assert reverse_transformed1['id2'].tolist() != reverse_transformed2['id2'].tolist()

        # Run - make sure resetting randomization works
        ht.reset_randomization()
        transformed = ht.transform(data)
        reverse_transformed3 = ht.reverse_transform(transformed)

        # Assert
        pd.testing.assert_frame_equal(reverse_transformed1, reverse_transformed3)

    def test_anonymized_faker_text(self):
        """Test ``AnonymizedFaker`` with text column."""
        # Setup
        data = pd.DataFrame({
            'id1': ['a', 'b', 'c'],
            'info': ['d', 'e', 'f'],
        })
        ht = HyperTransformer()

        # Run - simple run
        ht.detect_initial_config(data)
        ht.update_sdtypes({
            'id1': 'pii',
            'info': 'text'
        })
        ht.update_transformers({
            'id1': AnonymizedFaker(),
            'info': AnonymizedFaker()
        })
        ht.fit(data)
        transformed = ht.transform(data)
        reverse_transformed = ht.reverse_transform(transformed)

        # Assert
        assert all(reverse_transformed['info'].apply(lambda s: isinstance(s, str)))

    def test_pseudo_anonymized_faker(self):
        """Test ``PseudoAnonymizedFaker`` generates different values for different columns."""
        # Setup
        data = pd.DataFrame({
            'id1': ['a', 'b', 'c'],
            'id2': ['a', 'b', 'c'],
        })
        ht = HyperTransformer()

        # Run
        ht.detect_initial_config(data)
        ht.update_sdtypes({
            'id1': 'pii',
            'id2': 'pii'
        })
        ht.update_transformers({
            'id1': PseudoAnonymizedFaker(),
            'id2': PseudoAnonymizedFaker()
        })
        ht.fit(data)
        transformed = ht.transform(data)
        reverse_transformed1 = ht.reverse_transform(transformed)

        # Assert
        assert reverse_transformed1['id1'].tolist() != reverse_transformed1['id2'].tolist()

        # Run - run it again on the exact same data
        ht = HyperTransformer()
        ht.detect_initial_config(data)
        ht.update_sdtypes({
            'id1': 'pii',
            'id2': 'pii'
        })
        ht.update_transformers({
            'id1': PseudoAnonymizedFaker(),
            'id2': PseudoAnonymizedFaker()
        })
        ht.fit(data)
        transformed = ht.transform(data)
        reverse_transformed2 = ht.reverse_transform(transformed)

        # Assert - different instances of the same transformer should return the same result
        assert reverse_transformed1['id1'].tolist() == reverse_transformed2['id1'].tolist()

    def test_anonymized_faker_different_tables(self):
        """Test ``AnonymizedFaker`` generates different values for columns with same name."""
        # Setup
        data1 = pd.DataFrame({
            'id1': ['a', 'b', 'c'],
            'id2': ['a', 'b', 'c'],
        })
        data2 = pd.DataFrame({
            'id1': ['d', 'e', 'f'],
            'id2': ['d', 'e', 'f'],
        })
        ht = HyperTransformer()

        # Run on data1
        ht.detect_initial_config(data1)
        ht.update_sdtypes({
            'id1': 'pii',
            'id2': 'pii'
        })
        ht.update_transformers({
            'id1': AnonymizedFaker(),
            'id2': PseudoAnonymizedFaker()
        })
        ht.fit(data1)
        transformed = ht.transform(data1)
        reverse_transformed1 = ht.reverse_transform(transformed)

        # Run on data2
        ht.detect_initial_config(data2)
        ht.update_sdtypes({
            'id1': 'pii',
            'id2': 'pii'
        })
        ht.update_transformers({
            'id1': AnonymizedFaker(),
            'id2': PseudoAnonymizedFaker()
        })
        ht.fit(data2)
        transformed = ht.transform(data2)
        reverse_transformed2 = ht.reverse_transform(transformed)

        # Assert
        assert reverse_transformed1['id1'].tolist() != reverse_transformed2['id1'].tolist()
        assert reverse_transformed1['id2'].tolist() != reverse_transformed2['id2'].tolist()

    def test_random_seed(self):
        # Setup
        data = pd.DataFrame({
            'num1': [1, np.nan, 2] * 10,
            'num2': [1, np.nan, 2] * 10,
            'num3': [1, np.nan, 2] * 10,
            'num4': [1, np.nan, 2] * 10,
            'num5': [1, np.nan, 2] * 10,
            'num6': [1, np.nan, 2] * 10,
            'date1': [np.datetime64('2020-10-10'), np.datetime64('2021-11-11'), np.nan] * 10,
            'date2': [np.datetime64('2020-10-10'), np.datetime64('2021-11-11'), np.nan] * 10,
            'date3': [np.datetime64('2020-10-10'), np.datetime64('2021-11-11'), np.nan] * 10,
            'date4': [np.datetime64('2020-10-10'), np.datetime64('2021-11-11'), np.nan] * 10,
        })

        ht = HyperTransformer()
        ht.detect_initial_config(data)
        ht.update_transformers({
            'num1': FloatFormatter(),
            'num2': FloatFormatter(),
            'num3': ClusterBasedNormalizer(),
            'num4': ClusterBasedNormalizer(),
            'num5': GaussianNormalizer(),
            'num6': GaussianNormalizer(),
            'date1': UnixTimestampEncoder(),
            'date2': UnixTimestampEncoder(),
            'date3': OptimizedTimestampEncoder(),
            'date4': OptimizedTimestampEncoder(),
        })

        # Run
        ht.fit(data)
        transformed = ht.transform(data)
        reversed1 = ht.reverse_transform(transformed)

        # Assert
        assert reversed1['num1'].isna().tolist() != reversed1['num2'].isna().tolist()
        assert reversed1['num3'].isna().tolist() != reversed1['num4'].isna().tolist()
        assert reversed1['num5'].isna().tolist() != reversed1['num6'].isna().tolist()
        assert reversed1['date1'].isna().tolist() != reversed1['date2'].isna().tolist()
        assert reversed1['date3'].isna().tolist() != reversed1['date4'].isna().tolist()

        # Run
        ht.reset_randomization()
        transformed = ht.transform(data)
        reversed2 = ht.reverse_transform(transformed)

        # Assert
        pd.testing.assert_frame_equal(reversed1, reversed2)



import pickle

import numpy as np
import pandas as pd

from rdt.transformers.text import IDGenerator, RegexGenerator


class TestIDGenerator():

    def test_end_to_end(self):
        """End to end test of the ``IDGenerator``."""
        # Setup
        data = pd.DataFrame({
            'id': [1, 2, 3, 4, 5],
            'username': ['a', 'b', 'c', 'd', 'e']
        })

        # Run
        transformer = IDGenerator(prefix='id_', starting_value=100, suffix='_X')
        transformed = transformer.fit_transform(data, 'id')
        reverse_transform = transformer.reverse_transform(transformed)
        reverse_transform_2 = transformer.reverse_transform(transformed)
        transformer.reset_randomization()
        reverse_transform_3 = transformer.reverse_transform(transformed)

        # Assert
        expected_transformed = pd.DataFrame({
            'username': ['a', 'b', 'c', 'd', 'e']
        })

        expected_reverse_transform = pd.DataFrame({
            'username': ['a', 'b', 'c', 'd', 'e'],
            'id': ['id_100_X', 'id_101_X', 'id_102_X', 'id_103_X', 'id_104_X']
        })

        expected_reverse_transform_2 = pd.DataFrame({
            'username': ['a', 'b', 'c', 'd', 'e'],
            'id': ['id_105_X', 'id_106_X', 'id_107_X', 'id_108_X', 'id_109_X']
        })

        pd.testing.assert_frame_equal(transformed, expected_transformed)
        pd.testing.assert_frame_equal(reverse_transform, expected_reverse_transform)
        pd.testing.assert_frame_equal(reverse_transform_2, expected_reverse_transform_2)
        pd.testing.assert_frame_equal(reverse_transform_3, expected_reverse_transform)


def test_regexgenerator():
    """Test ``RegexGenerator`` with the default parameters."""
    # Setup
    data = pd.DataFrame({
        'id': [1, 2, 3, 4, 5],
        'username': ['a', 'b', 'c', 'd', 'e']
    })

    # Run
    instance = RegexGenerator()
    transformed = instance.fit_transform(data, 'id')
    reverse_transform = instance.reverse_transform(transformed)

    # Assert
    expected_transformed = pd.DataFrame({
        'username': ['a', 'b', 'c', 'd', 'e']
    })
    expected_reverse_transformed = pd.DataFrame({
        'username': ['a', 'b', 'c', 'd', 'e'],
        'id': ['AAAAA', 'AAAAB', 'AAAAC', 'AAAAD', 'AAAAE'],
    })

    pd.testing.assert_frame_equal(transformed, expected_transformed)
    pd.testing.assert_frame_equal(reverse_transform, expected_reverse_transformed)


def test_regexgenerator_with_custom_regex():
    """Test the ``RegexGenerator`` with a custom regex format."""
    # Setup
    data = pd.DataFrame({
        'id': [1, 2, 3, 4, 5],
        'username': ['a', 'b', 'c', 'd', 'e'],
    })

    # Run
    instance = RegexGenerator(regex_format='[1-9]')
    transformed = instance.fit_transform(data, 'id')
    reverse_transform = instance.reverse_transform(transformed)

    # Assert
    expected_transformed = pd.DataFrame({
        'username': ['a', 'b', 'c', 'd', 'e'],
    })

    expected_reverse_transformed = pd.DataFrame({
        'username': ['a', 'b', 'c', 'd', 'e'],
        'id': ['1', '2', '3', '4', '5'],
    })

    pd.testing.assert_frame_equal(transformed, expected_transformed)
    pd.testing.assert_frame_equal(reverse_transform, expected_reverse_transformed)


def test_regexgenerator_with_nans():
    """Test the ``RegexGenerator`` with a custom regex format and ``nans``."""
    # Setup
    data = pd.DataFrame({
        'id': [1, 2, 3, 4, 5],
        'username': ['a', np.nan, 'c', 'd', 'e']
    })

    # Run
    instance = RegexGenerator('[A-Z]')
    transformed = instance.fit_transform(data, 'username')
    reverse_transform = instance.reverse_transform(transformed)

    # Assert
    expected_transformed = pd.DataFrame({
        'id': [1, 2, 3, 4, 5],
    })

    expected_reverse_transformed = pd.DataFrame({
        'id': [1, 2, 3, 4, 5],
        'username': ['A', 'B', 'C', 'D', 'E'],
    })

    pd.testing.assert_frame_equal(transformed, expected_transformed)
    pd.testing.assert_frame_equal(reverse_transform, expected_reverse_transformed)


def test_regexgenerator_data_length_bigger_than_regex():
    """Test the ``RegexGenerator`` with short regex and more data length."""
    # Setup
    data = pd.DataFrame({
        'id': [1, 2, 3, 4, 5],
        'username': ['a', np.nan, 'c', 'd', 'e']
    })

    # Run
    instance = RegexGenerator('[a-b]')
    transformed = instance.fit_transform(data, 'username')
    reverse_transform = instance.reverse_transform(transformed)

    # Assert
    expected_transformed = pd.DataFrame({
        'id': [1, 2, 3, 4, 5],
    })

    expected_reverse_transformed = pd.DataFrame({
        'id': [1, 2, 3, 4, 5],
        'username': ['a', 'b', 'a', 'b', 'a'],
    })

    pd.testing.assert_frame_equal(transformed, expected_transformed)
    pd.testing.assert_frame_equal(reverse_transform, expected_reverse_transformed)


def test_regexgenerator_input_data_bigger_than_data_length():
    """Test the ``RegexGenerator`` with input dataframe bigger than the learned data length."""
    # Setup
    data = pd.DataFrame({
        'id': [1, 2, 3, 4, 5],
        'username': ['a', 'b', 'c', 'd', 'e']
    })

    # Run
    instance = RegexGenerator('[a-b]')
    instance.fit(data, 'username')

    transformed = pd.DataFrame({
        'id': [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
    })
    reverse_transform = instance.reverse_transform(transformed)

    # Assert
    expected_reverse_transformed = pd.DataFrame({
        'id': [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
        'username': ['a', 'b', 'a', 'b', 'a', 'b', 'a', 'b', 'a', 'b'],
    })

    pd.testing.assert_frame_equal(reverse_transform, expected_reverse_transformed)


def test_regexgenerator_called_multiple_times():
    """Test the ``RegexGenerator`` with short regex and called multiple times.

    This test ensures that when ``enforce_uniqueness`` is ``False`` this generator will continue
    to work.
    """
    # Setup
    data = pd.DataFrame({
        'id': [1, 2, 3, 4, 5],
        'username': ['a', np.nan, 'c', 'd', 'e']
    })

    instance = RegexGenerator('[a-c]')

    # Transform
    transformed = instance.fit_transform(data, 'username')

    # Assert Transform
    expected_transformed = pd.DataFrame({
        'id': [1, 2, 3, 4, 5],
    })
    pd.testing.assert_frame_equal(transformed, expected_transformed)

    # Reverse Transform
    first_reverse_transform = instance.reverse_transform(transformed)

    # Assert Reverse Transform
    expected_reverse_transformed = pd.DataFrame({
        'id': [1, 2, 3, 4, 5],
        'username': ['a', 'b', 'c', 'a', 'b'],
    })
    pd.testing.assert_frame_equal(first_reverse_transform, expected_reverse_transformed)

    # Reverse Transform Again
    second_reverse_transform = instance.reverse_transform(transformed.head(1))

    # Assert Reverse Transform
    expected_reverse_transformed = pd.DataFrame({
        'id': [1],
        'username': ['a'],
    })
    pd.testing.assert_frame_equal(second_reverse_transform, expected_reverse_transformed)

    # Reverse Transform Again
    third_reverse_transform = instance.reverse_transform(transformed.head(1))

    # Assert Reverse Transform
    expected_reverse_transformed = pd.DataFrame({
        'id': [1],
        'username': ['b'],
    })
    pd.testing.assert_frame_equal(third_reverse_transform, expected_reverse_transformed)


def test_regexgenerator_called_multiple_times_enforce_uniqueness():
    """Test that calling multiple times with ``enforce_uniqueness`` returns unique values."""
    # Setup
    data = pd.DataFrame({'my_column': np.arange(10)})
    generator = RegexGenerator(enforce_uniqueness=True)

    # Run
    transformed_data = generator.fit_transform(data, 'my_column')
    first_reverse_transform = generator.reverse_transform(transformed_data.head(3))
    second_reverse_transform = generator.reverse_transform(transformed_data.head(5))

    # Assert
    expected_first_reverse_transform = pd.DataFrame({
        'my_column': ['AAAAA', 'AAAAB', 'AAAAC']
    })
    expected_second_reverse_transform = pd.DataFrame({
        'my_column': ['AAAAD', 'AAAAE', 'AAAAF', 'AAAAG', 'AAAAH']
    })
    pd.testing.assert_frame_equal(first_reverse_transform, expected_first_reverse_transform)
    pd.testing.assert_frame_equal(second_reverse_transform, expected_second_reverse_transform)


def test_regexgenerator_pickled(tmpdir):
    """Test that ensures that ``RegexGenerator`` can be pickled."""
    # Setup
    data = pd.DataFrame({
        'id': [1, 2, 3, 4, 5],
        'username': ['a', 'b', 'c', 'd', 'e']
    })

    # Run
    instance = RegexGenerator()
    transformed = instance.fit_transform(data, 'id')
    instance.reverse_transform(transformed)

    # Pickle
    with open(tmpdir / 'file.pkl', 'wb') as f:
        pickle.dump(instance, f)

    with open(tmpdir / 'file.pkl', 'rb') as f:
        loaded = pickle.load(f)

    # Assert
    assert next(instance.generator) == next(loaded.generator)


def test_regexgenerator_with_many_possibilities():
    """Test the ``RegexGenerator`` with regex containing many possibilities."""
    # Setup
    data = pd.DataFrame({
        'id': ['a' * 50, 'a' * 49 + 'b', 'a' * 49 + 'c', 'a' * 49 + 'd', 'a' * 49 + 'e'],
        'username': ['aa', 'bb', 'cc', 'dd', 'ee'],
    })

    # Run
    instance = RegexGenerator(regex_format='[a-z]{50}')
    transformed = instance.fit_transform(data, 'id')
    reverse_transform = instance.reverse_transform(transformed)

    # Assert
    expected_transformed = pd.DataFrame({
        'username': ['aa', 'bb', 'cc', 'dd', 'ee'],
    })

    expected_reverse_transformed = pd.DataFrame({
        'username': ['aa', 'bb', 'cc', 'dd', 'ee'],
        'id': ['a' * 50, 'a' * 49 + 'b', 'a' * 49 + 'c', 'a' * 49 + 'd', 'a' * 49 + 'e'],
    })

    pd.testing.assert_frame_equal(transformed, expected_transformed)
    pd.testing.assert_frame_equal(reverse_transform, expected_reverse_transformed)



import re

import numpy as np
import pandas as pd
import pytest

from rdt.errors import TransformerProcessingError
from rdt.transformers.pii import AnonymizedFaker, PseudoAnonymizedFaker


class TestAnonymizedFaker:
    def test_default_settings(self):
        """End to end test with the default settings of the ``AnonymizedFaker``."""
        data = pd.DataFrame({
            'id': [1, 2, 3, 4, 5],
            'username': ['a', 'b', 'c', 'd', 'e']
        })

        instance = AnonymizedFaker()
        transformed = instance.fit_transform(data, 'username')

        reverse_transform = instance.reverse_transform(transformed)
        expected_transformed = pd.DataFrame({
            'id': [1, 2, 3, 4, 5]
        })

        pd.testing.assert_frame_equal(transformed, expected_transformed)
        assert len(reverse_transform['username']) == 5

    def test_get_supported_sdtypes(self):
        """Test that the correct supported sdtypes are returned."""
        # Run
        supported_sdtypes = AnonymizedFaker.get_supported_sdtypes()

        # Assert
        assert sorted(supported_sdtypes) == sorted(['pii', 'text'])

    def test_custom_provider(self):
        """End to end test with a custom provider and function for the ``AnonymizedFaker``."""
        data = pd.DataFrame({
            'id': [1, 2, 3, 4, 5],
            'username': ['a', 'b', 'c', 'd', 'e'],
            'cc': [
                '2276346007210438',
                '4149498289355',
                '213144860944676',
                '4514775286178',
                '213133122335401'
            ]
        })

        instance = AnonymizedFaker('credit_card', 'credit_card_number')
        transformed = instance.fit_transform(data, 'cc')
        reverse_transform = instance.reverse_transform(transformed)

        expected_transformed = pd.DataFrame({
            'id': [1, 2, 3, 4, 5],
            'username': ['a', 'b', 'c', 'd', 'e'],
        })

        pd.testing.assert_frame_equal(transformed, expected_transformed)
        assert len(reverse_transform['cc']) == 5

    def test_with_nans(self):
        """Test with the default settings of the ``AnonymizedFaker`` with ``nan`` values."""
        data = pd.DataFrame({
            'id': [1, 2, 3, 4, 5],
            'username': ['a', np.nan, 'c', 'd', 'e']
        })

        instance = AnonymizedFaker()
        transformed = instance.fit_transform(data, 'username')
        reverse_transform = instance.reverse_transform(transformed)

        expected_transformed = pd.DataFrame({
            'id': [1, 2, 3, 4, 5],
        })

        pd.testing.assert_frame_equal(transformed, expected_transformed)
        assert len(reverse_transform['username']) == 5
        assert reverse_transform['username'].isna().sum() == 1

    def test_with_nans_missing_value_generation_none(self):
        """End to end test settings missing_value_generation=None."""
        data = pd.DataFrame({
            'id': [1, 2, 3, 4, 5],
            'username': ['a', np.nan, 'c', 'd', 'e']
        })

        instance = AnonymizedFaker(missing_value_generation=None)
        transformed = instance.fit_transform(data, 'username')
        reverse_transform = instance.reverse_transform(transformed)

        expected_transformed = pd.DataFrame({
            'id': [1, 2, 3, 4, 5],
        })

        pd.testing.assert_frame_equal(transformed, expected_transformed)
        assert len(reverse_transform['username']) == 5
        assert reverse_transform['username'].isna().sum() == 0

    def test_custom_provider_with_nans(self):
        """End to end test with a custom provider for the ``AnonymizedFaker`` with `` nans``."""
        data = pd.DataFrame({
            'id': [1, 2, 3, 4, 5],
            'username': ['a', 'b', 'c', 'd', 'e'],
            'cc': [
                '2276346007210438',
                np.nan,
                '213144860944676',
                '4514775286178',
                '213133122335401'
            ]
        })

        instance = AnonymizedFaker(
            'credit_card',
            'credit_card_number',
        )
        transformed = instance.fit_transform(data, 'cc')
        reverse_transform = instance.reverse_transform(transformed)

        expected_transformed = pd.DataFrame({
            'id': [1, 2, 3, 4, 5],
            'username': ['a', 'b', 'c', 'd', 'e'],
        })

        pd.testing.assert_frame_equal(transformed, expected_transformed)
        assert len(reverse_transform['cc']) == 5
        assert reverse_transform['cc'].isna().sum() == 1

    def test_enforce_uniqueness(self):
        """Test that ``AnonymizedFaker`` works with uniqueness.

        Also ensure that when we call ``reset_randomization`` the generator will be able to
        create values again.
        """
        data = pd.DataFrame({
            'job': np.arange(500)
        })

        instance = AnonymizedFaker('job', 'job', enforce_uniqueness=True)
        transformed = instance.fit_transform(data, 'job')
        reverse_transform = instance.reverse_transform(transformed)

        # Assert
        assert len(reverse_transform['job'].unique()) == 500

        error_msg = re.escape(
            'The Faker function you specified is not able to generate 500 unique '
            'values. Please use a different Faker function for column '
            "('job')."
        )
        with pytest.raises(TransformerProcessingError, match=error_msg):
            instance.reverse_transform(transformed)

        instance.reset_randomization()
        instance.reverse_transform(transformed)


class TestPsuedoAnonymizedFaker:
    def test_default_settings(self):
        """End to end test with the default settings of the ``PseudoAnonymizedFaker``."""
        data = pd.DataFrame({
            'animals': ['cat', 'dog', 'parrot', 'monkey']
        })

        instance = PseudoAnonymizedFaker()

        transformed = instance.fit_transform(data, 'animals')
        reverse_transformed = instance.reverse_transform(transformed)

        assert transformed.columns == ['animals']
        pd.testing.assert_series_equal(
            reverse_transformed['animals'].map(instance._reverse_mapping_dict),
            data['animals']
        )
        unique_animals = set(reverse_transformed['animals'])
        assert unique_animals.intersection(set(instance._mapping_dict)) == set()
        assert len(reverse_transformed) == len(transformed) == 4

    def test_with_nans(self):
        """Test with the default settings of the ``PseudoAnonymizedFaker`` and ``nans``."""
        data = pd.DataFrame({
            'animals': ['cat', 'dog', np.nan, 'monkey']
        })

        instance = PseudoAnonymizedFaker()

        transformed = instance.fit_transform(data, 'animals')
        reverse_transformed = instance.reverse_transform(transformed)

        assert transformed.columns == ['animals']
        pd.testing.assert_series_equal(
            reverse_transformed['animals'].map(instance._reverse_mapping_dict),
            data['animals']
        )
        unique_animals = set(reverse_transformed['animals'])
        assert unique_animals.intersection(set(instance._mapping_dict)) == set()
        assert len(reverse_transformed) == len(transformed) == 4

    def test_with_custom_provider(self):
        """End to end test with custom settings of the ``PseudoAnonymizedFaker``."""
        data = pd.DataFrame({
            'animals': ['cat', 'dog', np.nan, 'monkey']
        })

        instance = PseudoAnonymizedFaker('credit_card', 'credit_card_number')

        transformed = instance.fit_transform(data, 'animals')
        reverse_transformed = instance.reverse_transform(transformed)

        assert transformed.columns == ['animals']
        pd.testing.assert_series_equal(
            reverse_transformed['animals'].map(instance._reverse_mapping_dict),
            data['animals']
        )
        unique_animals = set(reverse_transformed['animals'])
        assert unique_animals.intersection(set(instance._mapping_dict)) == set()
        assert len(reverse_transformed) == len(transformed) == 4



"""Test Personal Identifiable Information Transformer using Faker."""

import pickle
import re
import tempfile
from unittest.mock import Mock, call, patch

import numpy as np
import pandas as pd
import pytest

from rdt.errors import TransformerInputError, TransformerProcessingError
from rdt.transformers.categorical import LabelEncoder
from rdt.transformers.pii.anonymizer import AnonymizedFaker, PseudoAnonymizedFaker


class TestAnonymizedFaker:
    """Test class for ``AnonymizedFaker``."""

    @patch('rdt.transformers.pii.anonymizer.faker')
    @patch('rdt.transformers.pii.anonymizer.getattr')
    @patch('rdt.transformers.pii.anonymizer.attrgetter')
    def test_check_provider_function_baseprovider(self, mock_attrgetter, mock_getattr, mock_faker):
        """Test that ``getattr`` is being called with ``BaseProvider`` and ``function_name``.

        Mock:
            - Mock the ``getattr`` from Python to ensure that is being called with the input.
            - Mock faker and ensure that ``getattr`` is being called with ``faker.providers``.
        """
        # Setup
        mock_attrgetter.return_value = lambda x: 'module'
        mock_getattr.side_effect = ['provider', None]

        # Run
        AnonymizedFaker.check_provider_function('BaseProvider', 'function_name')

        # Assert
        assert mock_attrgetter.call_args_list[0] == call('BaseProvider')
        assert mock_getattr.call_args_list[0] == call('module', 'function_name')

    @patch('rdt.transformers.pii.anonymizer.faker')
    @patch('rdt.transformers.pii.anonymizer.getattr')
    @patch('rdt.transformers.pii.anonymizer.attrgetter')
    def test_check_provider_function_other_providers(self, mock_attrgetter, mock_getattr,
                                                     mock_faker):
        """Test that ``getattr`` is being called with ``provider_name`` and ``function_name``.

        Mock:
            - Mock the ``getattr`` from Python to ensure that is being called with the input.
            - Mock faker and ensure that ``getattr`` is being called with ``faker.providers``.
        """
        # Setup
        mock_attrgetter.return_value = lambda x: 'module'
        mock_getattr.side_effect = ['provider_class', None]

        # Run
        AnonymizedFaker.check_provider_function('provider_name', 'function_name')

        # Assert
        assert mock_attrgetter.call_args_list[0] == call('provider_name')
        assert mock_getattr.call_args_list[0] == call('module', 'Provider')
        assert mock_getattr.call_args_list[1] == call('provider_class', 'function_name')

    def test_check_provider_function_raise_attribute_error(self):
        """Test that ``check_provider_function`` raises an ``AttributeError``.

        Test when the provider class is not in the ``faker.providers`` an ``AttributeError`` is
        being raised or when the ``function`` is not within the ``provider``.
        """
        # Setup
        expected_message = (
            "The 'TestProvider' module does not contain a function named "
            "'TestFunction'.\nRefer to the Faker docs to find the correct function: "
            'https://faker.readthedocs.io/en/master/providers.html'
        )

        # Run
        with pytest.raises(TransformerProcessingError, match=expected_message):
            AnonymizedFaker.check_provider_function('TestProvider', 'TestFunction')

    def test__function_enforce_uniqueness_false(self):
        """Test that ``_function`` does not use ``faker.unique``.

        The method ``_function`` should return a call from the
        ``instance.faker.provider.<function>``.

        Mock:
            - Instance of 'AnonymizedFaker'.
            - Instance ``enforce_uniqueness`` set to `False`
            - Faker instance.
            - A function for the faker instance.

        Output:
            - Return value of mocked function.

        Side Effects:
            - The returned function, when called, has to call the `faker.<function_name>` function
              with the provided kwargs.
        """
        # setup
        instance = Mock()
        instance.enforce_uniqueness = False
        function = Mock()
        unique_function = Mock()
        function.return_value = 1

        instance.faker.unique.number = unique_function
        instance.faker.number = function
        instance.function_name = 'number'
        instance.function_kwargs = {'type': 'int'}

        # Run
        result = AnonymizedFaker._function(instance)

        # Assert
        unique_function.assert_not_called()
        function.assert_called_once_with(type='int')
        assert result == 1

    def test__function_enforce_uniqueness_true(self):
        """Test that ``_function`` uses the ``faker.unique``.

        The method ``_function`` should return a call from the
        ``instance.faker.unique.<function>``.

        Mock:
            - Instance of 'AnonymizedFaker'.
            - Instance ``enforce_uniqueness`` set to ``True``
            - Faker instance.
            - A function for the faker instance.

        Output:
            - Return value of mocked function.

        Side Effects:
            - The returned function, when called, has to call the ``faker.unique.<function_name>``
              function with the provided kwargs.
        """
        # setup
        instance = Mock()
        instance.enforce_uniqueness = True
        function = Mock()
        unique_function = Mock()
        unique_function.return_value = 1

        instance.faker.unique.number = unique_function
        instance.faker.number = function
        instance.function_name = 'number'
        instance.function_kwargs = {'type': 'int'}

        # Run
        result = AnonymizedFaker._function(instance)

        # Assert
        function.assert_not_called()
        unique_function.assert_called_once_with(type='int')
        assert result == 1

    @patch('rdt.transformers.pii.anonymizer.importlib')
    @patch('rdt.transformers.pii.anonymizer.warnings')
    def test__check_locales(self, mock_warnings, mock_importlib):
        """Test that check locales warns the user if the spec was not found.

        Mock:
            - Mock importlib with side effects to return one `None` and one value.
            - Mock the warnings.
        Side Effect:
            - mock_warnings has been called once with the expected message.
        """
        # Setup
        instance = Mock()
        instance.provider_name = 'credit_card'
        instance.function_name = 'credit_card_full'
        instance.locales = ['es_ES', 'en_US']
        mock_importlib.util.find_spec.side_effect = [None, 'en_US']

        # Run
        AnonymizedFaker._check_locales(instance)

        # Assert
        expected_message = (
            "Locales ['es_ES'] do not support provider 'credit_card' "
            "and function 'credit_card_full'.\nIn place of these locales, 'en_US' will "
            'be used instead. Please refer to the localized provider docs for more '
            'information: https://faker.readthedocs.io/en/master/locales.html'
        )
        mock_warnings.warn.assert_called_once_with(expected_message)

    @patch('rdt.transformers.pii.anonymizer.faker')
    @patch('rdt.transformers.pii.anonymizer.AnonymizedFaker.check_provider_function')
    def test___init__default(self, mock_check_provider_function, mock_faker):
        """Test the default instantiation of the transformer.

        Test that by default the transformer is being instantiated with ``BaseProvider`` and
        ``lexify``.

        Mock:
            - ``check_provider_function`` mock and assert that this is being called with
              ``BaseProvider`` and ``lexify``.
            - ``faker`` mock to ensure that is being called with ``None`` as locales
              and that the ``lexify`` from the instance is being assigned to the ``_function``.

        Side effects:
            - the ``instance.provider_name`` is ``BaseProvider``.
            - the ``instance.function_name`` is ``lexify``.
            - the ``instance.locales`` is ``None``.
            - the ``instance.function_kwargs`` is an empty ``dict``.
            - ``check_provider_function`` has been called once with ``BaseProvider`` and
              ``lexify``.
            - the ``instance._function`` is ``instance.faker.lexify``.
        """
        # Run
        instance = AnonymizedFaker()

        # Assert
        mock_check_provider_function.assert_called_once_with('BaseProvider', 'lexify')
        assert instance.provider_name == 'BaseProvider'
        assert instance.function_name == 'lexify'
        assert instance.function_kwargs == {}
        assert instance.locales is None
        assert mock_faker.Faker.called_once_with(None)
        assert instance.enforce_uniqueness is False
        assert instance.missing_value_generation == 'random'

    def test___init__error_missing_value_generation(self):
        """Test that the ``__init__`` raises an error if the missing value generation is invalid.

        Test that the ``__init__`` raises an error if the missing value generation is invalid.

        Side effects:
            - ``ValueError`` is raised.
        """
        expected_message = (
            "Missing value generation 'invalid' is not supported "
            "for AnonymizedFaker. Please use either 'random' or None."
        )
        # Run and Assert
        with pytest.raises(TransformerInputError, match=expected_message):
            AnonymizedFaker(missing_value_generation='invalid')

    @patch('rdt.transformers.pii.anonymizer.faker')
    @patch('rdt.transformers.pii.anonymizer.AnonymizedFaker.check_provider_function')
    def test___init__custom(self, mock_check_provider_function, mock_faker):
        """Test the instantiation of the transformer with custom parameters.

        Test that the transformer can be instantiated with a custom provider and function, and
        this is being assigned properly.

        Mock:
            - ``check_provider_function`` mock and assert that this is being called with
              ``CreditCard`` and ``credit_card_full``.
            - ``faker`` mock to ensure that is being called with ``['en_US', 'fr_FR']`` as locales
              and that the ``credit_card_full`` from the instance is being assigned to the
              ``_function``.

        Side effects:
            - the ``instance.provider_name`` is ``CreditCard``.
            - the ``instance.function_name`` is ``credit_card_full``.
            - the ``instance.locales`` is ``['en_US', 'fr_FR']``.
            - the ``instance.function_kwargs`` is ``{'type': 'visa'}``.
            - ``check_provider_function`` has been called once with ``CreditCard`` and
              ``credit_card_full``.
            - the ``instance._function`` is ``instance.faker.credit_card_full``.
        """
        # Run
        instance = AnonymizedFaker(
            provider_name='credit_card',
            function_name='credit_card_full',
            function_kwargs={
                'type': 'visa'
            },
            locales=['en_US', 'fr_FR'],
            enforce_uniqueness=True
        )

        # Assert
        mock_check_provider_function.assert_called_once_with('credit_card', 'credit_card_full')
        assert instance.provider_name == 'credit_card'
        assert instance.function_name == 'credit_card_full'
        assert instance.function_kwargs == {'type': 'visa'}
        assert instance.locales == ['en_US', 'fr_FR']
        assert mock_faker.Faker.called_once_with(['en_US', 'fr_FR'])
        assert instance.enforce_uniqueness

    def test___init__no_function_name(self):
        """Test the instantiation of the transformer with custom parameters.

        Test that the transformer raises an error when no function name is provided.

        Raises:
            - Error.
        """
        # Run / Assert
        expected_message = (
            'Please specify the function name to use from the '
            "'credit_card' provider."
        )
        with pytest.raises(TransformerInputError, match=expected_message):
            AnonymizedFaker(provider_name='credit_card', locales=['en_US', 'fr_FR'])

    @patch('rdt.transformers.pii.anonymizer.issubclass')
    @patch('rdt.transformers.pii.anonymizer.BaseTransformer')
    def test_get_supported_sdtypes(self, base_mock, issubclass_mock):
        """Test that the method returns all sdtypes except the basic ones."""
        # Setup
        issubclass_mock.return_value = False
        numerical_mock = Mock()
        numerical_mock.get_supported_sdtypes.return_value = ['numerical']
        categorical_mock = Mock()
        categorical_mock.get_supported_sdtypes.return_value = ['categorical']
        datetime_mock = Mock()
        datetime_mock.get_supported_sdtypes.return_value = ['datetime']
        boolean_mock = Mock()
        boolean_mock.get_supported_sdtypes.return_value = ['boolean', 'categorical']
        text_mock = Mock()
        text_mock.get_supported_sdtypes.return_value = ['text']
        phone_mock = Mock()
        phone_mock.get_supported_sdtypes.return_value = ['phone_number']
        pii_mock = Mock()
        pii_mock.get_supported_sdtypes.return_value = ['pii']
        base_mock.get_subclasses.return_value = [
            numerical_mock,
            categorical_mock,
            datetime_mock,
            boolean_mock,
            text_mock,
            phone_mock,
            pii_mock
        ]

        # Run
        supported_sdtypes = AnonymizedFaker.get_supported_sdtypes()

        # Assert
        assert sorted(supported_sdtypes) == sorted(['phone_number', 'pii', 'text'])

    @patch('rdt.transformers.pii.anonymizer.BaseTransformer.reset_randomization')
    @patch('rdt.transformers.pii.anonymizer.faker')
    def test_reset_randomization(self, mock_faker, mock_base_reset):
        """Test that this function creates a new faker instance."""
        # Setup
        instance = AnonymizedFaker()
        instance.locales = ['en_US']

        # Run
        AnonymizedFaker.reset_randomization(instance)

        # Assert
        assert mock_faker.Faker.called_once_with(['en_US'])
        mock_base_reset.assert_called_once()

    def test__fit(self):
        """Test the ``_fit`` method.

        Validate that the ``_fit`` method learns the size of the input data.

        Setup:
            - Initialize a ``AnonymizedFaker`` transformer.

        Input:
            - ``pd.Series`` containing 3 strings.

        Side Effects:
            - ``instance.data_length`` equals to the length of the input data.
        """
        # Setup
        transformer = AnonymizedFaker()
        columns_data = pd.Series(['1', '2', '3', None, np.nan])
        transformer.columns = ['col']

        # Run
        transformer._fit(columns_data)

        # Assert
        assert transformer.data_length == 5
        assert transformer.output_properties == {None: {'next_transformer': None}}
        assert transformer._nan_frequency == 0.4

    def test__transform(self):
        """Test the ``_transform`` method.

        Validate that the ``_transform`` method returns ``None``.

        Setup:
            - Initialize a ``AnonymizedFaker`` transformer.

        Input:
            - ``pd.Series`` with three values.

        Output:
            - ``None``.
        """
        # Setup
        columns_data = pd.Series([1, 2, 3])
        instance = AnonymizedFaker()

        # Run
        result = instance._transform(columns_data)

        # Assert
        assert result is None

    def test__reverse_transform(self):
        """Test the ``_reverse_transform`` method.

        Validate that the ``_reverse_transform`` method calls the ``instance._function`` with
        the ``instance.function_kwargs`` the ``instance.data_length`` amount of times.

        Setup:
            - Initialize a ``AnonymizedFaker`` transformer.

        Input:
            - ``None``.

        Output:
            - the output a ``numpy.array`` with the generated values from
              ``instance._function``.
        """
        # Setup
        instance = AnonymizedFaker()
        instance.data_length = 3
        function = Mock()
        function.side_effect = ['a', 'b', 'c']

        instance._function = function

        # Run
        result = instance._reverse_transform(None)

        # Assert
        assert function.call_args_list == [call(), call(), call()]
        np.testing.assert_array_equal(result, np.array(['a', 'b', 'c']))

    def test__reverse_transform_with_nans(self):
        """Test that ``_reverse_transform`` generates NaNs."""
        # Setup
        instance = AnonymizedFaker()
        instance.data_length = 4
        instance._nan_frequency = 0.25
        function = Mock()
        function.side_effect = ['a', 'b', 'c', 'd']

        instance._function = function

        # Run
        result = instance._reverse_transform(None)
        result = pd.Series(result)

        # Assert
        assert function.call_args_list == [call(), call(), call(), call()]
        assert instance.missing_value_generation == 'random'
        assert result.isna().sum() == 1

    def test__reverse_transform_not_enough_unique_values(self):
        """Test the ``_reverse_transform`` method.

        Test that when calling the ``_reverse_transform`` method and the ``instance._function`` is
        not generating enough unique values raises an error.

        Setup:
            -Instance of ``AnonymizedFaker``.

        Input:
            - ``pandas.Series`` representing a column.

        Side Effect:
            - Raises an error.
        """
        # Setup
        instance = AnonymizedFaker('misc', 'boolean', enforce_uniqueness=True)
        data = pd.Series(['a', 'b', 'c', 'd'])
        instance.columns = ['a']

        # Run / Assert
        error_msg = re.escape(
            'The Faker function you specified is not able to generate 4 unique '
            "values. Please use a different Faker function for column ('a')."
        )
        with pytest.raises(TransformerProcessingError, match=error_msg):
            instance._reverse_transform(data)

    def test__reverse_transform_size_is_length_of_data(self):
        """Test the ``_reverse_transform`` method.

        Validate that the ``_reverse_transform`` method calls the ``instance._function`` with
        the ``instance.function_kwargs`` the ``len(data)`` amount of times.

        Setup:
            - Initialize a ``AnonymizedFaker`` transformer.

        Input:
            - ``pd.Series`` with three values.

        Output:
            - the output a ``numpy.array`` with the generated values from
              ``instance._function``.
        """
        # Setup
        instance = AnonymizedFaker()
        data = pd.Series([1, 2, 3])
        instance.data_length = 0
        function = Mock()
        function.side_effect = ['a', 'b', 'c']

        instance._function = function

        # Run
        result = instance._reverse_transform(data)

        # Assert
        assert function.call_args_list == [call(), call(), call()]
        np.testing.assert_array_equal(result, np.array(['a', 'b', 'c']))

    def test___repr__default(self):
        """Test the ``__repr__`` method.

        With the default parameters should return only the ``function_name='lexify'`` as an
        starting argument for the ``AnonymizedFaker``.
        """
        # Setup
        instance = AnonymizedFaker()

        # Run
        res = repr(instance)

        # Assert
        expected_res = "AnonymizedFaker(function_name='lexify')"
        assert res == expected_res

    def test___repr__custom_provider(self):
        """Test the ``__repr__`` method.

        With the custom args the ``repr`` of the class should return all the non default
        arguments.
        """
        # Setup
        instance = AnonymizedFaker('credit_card', 'credit_card_full')

        # Run
        res = repr(instance)

        # Assert
        expected = "AnonymizedFaker(provider_name='credit_card', function_name='credit_card_full')"
        assert res == expected


class TestPseudoAnonymizedFaker:
    """Test class for ``PseudoAnonymizedFaker``."""

    @patch('rdt.transformers.pii.anonymizer.warnings')
    def test___getstate__(self, mock_warnings):
        """Test that when pickling the warning message is being triggered."""
        # Setup
        instance = PseudoAnonymizedFaker()

        expected_warning_msg = (
            'You are saving the mapping information, which includes the original data. '
            'Sharing this object with others will also give them access to the original data '
            'used with this transformer.'
        )

        # Run / Assert
        with tempfile.TemporaryFile() as tmp:
            pickle.dump(instance, tmp)

        mock_warnings.warn.assert_called_once_with(expected_warning_msg)

    @patch('rdt.transformers.pii.anonymizer.faker')
    @patch('rdt.transformers.pii.anonymizer.AnonymizedFaker.check_provider_function')
    def test___init__super_attrs(self, mock_check_provider_function, mock_faker):
        """Test that initializing an instance is calling properly the ``super`` class.

        Mock:
            - ``check_provider_function`` mock and assert that this is being called with
              ``BaseProvider`` and ``lexify``.
            - ``faker`` mock to ensure that is being called with ``None`` as locales
              and that the ``lexify`` from the instance is being assigned to the ``_function``.

        Side effects:
            - the ``instance.provider_name`` is ``BaseProvider``.
            - the ``instance.function_name`` is ``lexify``.
            - the ``instance.locales`` is ``None``.
            - the ``instance.function_kwargs`` is an empty ``dict``.
            - ``check_provider_function`` has been called once with ``BaseProvider`` and
              ``lexify``.
            - the ``instance._function`` is ``instance.faker.lexify``.
        """
        # Run
        instance = PseudoAnonymizedFaker()

        # Assert
        assert instance._mapping_dict == {}
        assert instance._reverse_mapping_dict == {}

        # Assert Super Attrs
        assert instance.provider_name == 'BaseProvider'
        assert instance.function_name == 'lexify'
        assert instance.function_kwargs == {}
        assert instance.locales is None
        assert mock_faker.Faker.called_once_with(None)

    @patch('rdt.transformers.pii.anonymizer.faker')
    @patch('rdt.transformers.pii.anonymizer.AnonymizedFaker.check_provider_function')
    def test___init__custom(self, mock_check_provider_function, mock_faker):
        """Test the instantiation of the transformer with custom parameters.

        Test that the transformer can be instantiated with a custom provider and function, and
        this is being assigned properly.

        Mock:
            - ``check_provider_function`` mock and assert that this is being called with
              ``CreditCard`` and ``credit_card_full``.
            - ``faker`` mock to ensure that is being called with ``['en_US', 'fr_FR']`` as locales
              and that the ``credit_card_full`` from the instance is being assigned to the
              ``_function``.

        Side effects:
            - the ``instance.provider_name`` is ``CreditCard``.
            - the ``instance.function_name`` is ``credit_card_full``.
            - the ``instance.locales`` is ``['en_US', 'fr_FR']``.
            - the ``instance.function_kwargs`` is ``{'type': 'visa'}``.
            - ``check_provider_function`` has been called once with ``CreditCard`` and
              ``credit_card_full``.
            - the ``instance._function`` is ``instance.faker.credit_card_full``.
        """
        # Run
        instance = PseudoAnonymizedFaker(
            provider_name='credit_card',
            function_name='credit_card_full',
            function_kwargs={
                'type': 'visa'
            },
            locales=['en_US', 'fr_FR']
        )

        # Assert
        assert instance._mapping_dict == {}
        assert instance._reverse_mapping_dict == {}
        mock_check_provider_function.assert_called_once_with('credit_card', 'credit_card_full')
        assert instance.provider_name == 'credit_card'
        assert instance.function_name == 'credit_card_full'
        assert instance.function_kwargs == {'type': 'visa'}
        assert instance.locales == ['en_US', 'fr_FR']
        assert mock_faker.Faker.called_once_with(['en_US', 'fr_FR'])

    def test_get_mapping(self):
        """Test the ``get_mapping`` method.

        Validate that the ``get_mapping`` method return a ``deepcopy``  of the
        ``instance._mapping_dict``.

        Setup:
            - Instance of ``PseudoAnonymizedFaker``.
            - ``instance._mapping_dict`` to contain a dictionary.

        Output:
            - A deepcopy of the ``instance._mapping_dict``.
        """
        # Setup
        instance = PseudoAnonymizedFaker()
        instance._mapping_dict = {'a': 'b'}

        # Run
        result = instance.get_mapping()

        # Assert
        assert result == {'a': 'b'}
        assert id(result) != id(instance._mapping_dict)

    def test__fit(self):
        """Test the ``_fit`` method.

        Test that when calling the ``_fit`` method we are populating the ``instance._mapping_dict``
        with unique values and the ``instance._reverse_mapping_dict`` with those as well.

        Setup:
            -Instance of ``PseudoAnonymizedFaker``.

        Input:
            - ``pandas.Series`` representing a column.

        Mock:
            - Mock the ``instance._function`` to return controlled values.

        Side Effects:
            - ``instance._mapping_dict`` has been populated with the input unique data as keys and
              ``_function`` returned values as values.
            - ``instance._reverse_mapping_dict`` contains the ``_function`` returned values as keys
              and the input data as values.
        """
        # Setup
        instance = PseudoAnonymizedFaker()
        instance._function = Mock()
        instance._function.side_effect = [1, 2, 3]
        instance.columns = ['col']
        data = pd.Series(['a', 'b', 'c'])

        # Run
        instance._fit(data)

        # Assert
        assert instance._mapping_dict == {'a': 1, 'b': 2, 'c': 3}
        assert instance._reverse_mapping_dict == {1: 'a', 2: 'b', 3: 'c'}
        assert list(instance.output_properties) == [None]
        assert list(instance.output_properties[None]) == ['sdtype', 'next_transformer']
        assert instance.output_properties[None]['sdtype'] == 'categorical'

        transformer = instance.output_properties[None]['next_transformer']
        assert isinstance(transformer, LabelEncoder)
        assert transformer.add_noise is True

    def test__fit_not_enough_unique_values_in_faker_function(self):
        """Test the ``_fit`` method.

        Test that when calling the ``_fit`` method and the ``instance._function`` is not
        generating enough unique values raises an error.

        Setup:
            -Instance of ``PseudoAnonymizedFaker``.

        Input:
            - ``pandas.Series`` representing a column.

        Mock:
            - Mock the ``instance._function`` to return only 1 value.

        Side Effect:
            - Raises an error.
        """
        # Setup
        instance = PseudoAnonymizedFaker('misc', 'boolean')
        instance.columns = ['col']
        data = pd.Series(['a', 'b', 'c', 'd'])

        # Run / Assert
        error_msg = (
            'The Faker function you specified is not able to generate '
            '4 unique values. Please use a different '
            'Faker function for this column.'
        )
        with pytest.raises(TransformerProcessingError, match=error_msg):
            instance._fit(data)

    def test__transform(self):
        """Test the ``_transform`` method.

        Test that when ``_transform`` is being performed and no new values are present, the data
        is being mapped to the ``instance._mapping_dict``.

        Setup:
            - Instance of ``PseudoAnonymizedTransformer``.
            - ``_mapping_dict`` to the original data.

        Input:
            - pandas.Series with some data to be mapped.

        Output:
            - The ``pandas.Series`` with the ``mapped_data``.
        """
        # Setup
        instance = PseudoAnonymizedFaker()
        instance._mapping_dict = {'a': 'z', 'b': 'y', 'c': 'x'}
        instance.columns = ['col']

        data = pd.Series(['a', 'b', 'c'], name='col')

        # Run
        result = instance._transform(data)

        # Assert
        pd.testing.assert_series_equal(result, pd.Series(['z', 'y', 'x'], name='col'))

    def test__transform_with_new_values(self):
        """Test the ``_transform`` method.

        Test that when new data is being passed to the ``_transform``, this raises an error.

        Setup:
            - Instance of ``PseudoAnonymizedTransformer``.
            - ``_mapping_dict`` to the original data.

        Input:
            - pandas.Series with values that are not within the ``_mapping_dict``.

        Side Effects:
            - Raises a error.
        """
        # Setup
        instance = PseudoAnonymizedFaker()
        instance._mapping_dict = {'a': 'z', 'b': 'y', 'c': 'x'}

        # Assert / Run
        error_msg_short = re.escape(
            'The data you are transforming has new, unexpected values '
            '(1, 2, 3). Please fit the transformer again using this '
            'new data.'
        )
        error_msg_long = re.escape(
            'The data you are transforming has new, unexpected values '
            '(1, 2, 3, 4, 5 and 5 more). Please fit the transformer again using this '
            'new data.'
        )

        with pytest.raises(TransformerProcessingError, match=error_msg_short):
            instance._transform(pd.Series([1, 2, 3]))

        with pytest.raises(TransformerProcessingError, match=error_msg_long):
            instance._transform(pd.Series([1, 2, 3, 4, 5, 6, 7, 8, 9, 10]))

    def test__reverse_transform(self):
        """Test the ``_reverse_transform`` method.

        Test that the ``_reverse_transform`` returns the input data.

        Setup:
            - instance of ``PseudoAnonymizedFaker``.

        Input:
            - pd.Series

        Output:
            - The input data.
        """
        # Setup
        instance = PseudoAnonymizedFaker()
        instance.columns = ['col']

        data = pd.Series(['a', 'b', 'c'], name='col')

        # Run
        reverse_transformed = instance._reverse_transform(data)

        # Assert
        pd.testing.assert_series_equal(reverse_transformed, pd.Series(['a', 'b', 'c'], name='col'))
